package com.sitemesh.mapper;

import java.io.File;
import java.io.IOException;
import java.util.*;
import javax.servlet.Config;
import javax.servlet.ServletException;
import javax.servlet.http.HttpServletRequest;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import org.w3c.dom.*;
import org.xml.sax.SAXException;
import com.sitemesh.*;

// ---------------- PLEASE REFACTOR ME!!! --------------------

/**
 * Default implementation of DecoratorMapper. Reads decorators and
 * mappings from <code>/WEB-INF/decorators.xml</code>.
 *
 * @author <a href="joe@truemesh.com">Joe Walnes</a>
 * @author <a href="mcannon@internet.com">Mike Cannon-Brookes</a>
 * @version $Revision: 1.5 $
 *
 * @see com.sitemesh.DecoratorMapper
 * @see com.sitemesh.mapper.DefaultDecorator
 */
public class DefaultDecoratorMapper implements DecoratorMapper, RequestConstants {

	private Map decorators;
	private Map mappings;
	private long configLastModified;
	private File configFile;

	/**
	 * @label constructs
	 * @directed 
	 */
	/*#private DefaultDecorator lnkDefaultDecorator;*/

	public void init(Config config, Properties properties) throws InstantiationException {
		try {
			String fileName = config.getInitParameter("decorators.config");
			if (fileName == null) fileName = "/WEB-INF/decorators.xml";
			configFile = new File( config.getServletContext().getRealPath(fileName) );
			loadConfig();
		}
		catch (Exception e) {
			throw new InstantiationException(e.toString());
		}
	}

	public Decorator getDecorator(HttpServletRequest request, Page page) {
		Decorator result = null;
		if ( page instanceof HTMLPage ) {
			// Only try and find a Decorator if instance of HTMLPage
			if ( request.getAttribute(DECORATOR) != null ) {
				// Retrieve name of decorator to use from request
				String decoratorName = (String)request.getAttribute(DECORATOR);
				result = ( Decorator )decorators.get( decoratorName );
			}
			else {
				// If that doesn't work, get the decorator-mapping for the url.
				String key = findKey( request.getServletPath() );
				result = ( Decorator )decorators.get( ( String )mappings.get( key ) );
			}
		}
		return result;
	}

	private String findKey( String path ) {
		String result = findExactKey( path );
		if ( result == null ) result = findDirectoryKey( path );
		if ( result == null ) result = findExtentionKey( path );
		if ( result == null ) result = findDefaultKey( path );
		return result;
	}

	private String findExactKey( String path ) {
		if ( mappings.containsKey( path ) ) return path;
		return null;
	}

	private String findDirectoryKey( String path ) {
		if (path.lastIndexOf("/") <= 0) return null;
		String tempPath = path.substring(0, path.lastIndexOf("/"));
		while (!tempPath.equals("/") && tempPath.length() > 0) {
			if ( mappings.containsKey( tempPath + "/*" ) ) return tempPath + "/*";
			tempPath = tempPath.substring(0, tempPath.lastIndexOf("/"));
		}
		return null;
	}

	private String findExtentionKey( String path ) {
		if ( path.indexOf( "." ) == -1 ) return null;
		String ext = path.substring( path.lastIndexOf( "." ) + 1 );
		String[] extKeys = {"*."+ext , "/*."+ext};
		for ( int i = 0; i < extKeys.length; i++ ) {
			if ( mappings.containsKey( extKeys[i] ) ) return extKeys[i];
		}
		return null;
	}

	private String findDefaultKey( String path ) {
		String[] defaultKeys = {"/","*","/*"};
		for ( int i = 0; i < defaultKeys.length; i++ ) {
			if ( mappings.containsKey( defaultKeys[i] ) ) return defaultKeys[i];
		}
		return null;
	}

	private synchronized void loadConfig() throws ServletException {
		try {
			// Clear previous config
			mappings = new HashMap();
			decorators = new HashMap();
			// Parse file
			DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
			DocumentBuilder builder        = factory.newDocumentBuilder();
			configLastModified             = configFile.lastModified();
			Document document              = builder.parse( configFile );
			Element root                   = document.getDocumentElement();
			// Verify root element
			if ( !root.getTagName().equals( "decorators" ) ) {
				throw new ServletException( "Root element of decorators config file not <decorators>" );
			}
			// Get decorators
			NodeList decoratorNodes = root.getElementsByTagName( "decorator" );
			for ( int i = 0; i < decoratorNodes.getLength(); i++ ) {
				String name         = getContainedText( decoratorNodes.item( i ), "decorator-name" );
				String page         = getContainedText( decoratorNodes.item( i ), "jsp-file" );
				Map params          = new HashMap();
				NodeList paramNodes = ( ( Element )decoratorNodes.item( i ) ).getElementsByTagName( "init-param" );
				for ( int ii = 0; ii < paramNodes.getLength(); ii++ ) {
					String paramName = getContainedText( paramNodes.item( ii ), "param-name" );
					String paramValue = getContainedText( paramNodes.item( ii ), "param-value" );
					params.put( paramName, paramValue );
				}
				storeDecorator( new DefaultDecorator( name, page, params ) );
			}
			// Get decorator-mappings
			NodeList mappingNodes = root.getElementsByTagName( "decorator-mapping" );
			for ( int i = 0; i < mappingNodes.getLength(); i++ ) {
				Element n      = ( Element )mappingNodes.item( i );
				String name    = getContainedText( mappingNodes.item( i ), "decorator-name" );
				String pattern = getContainedText( mappingNodes.item( i ), "url-pattern" );
				storeMapping( name, pattern );
			}
		}
		catch ( ParserConfigurationException e ) {
			throw new ServletException( "Could not get XML parser", e );
		}
		catch ( IOException e ) {
			throw new ServletException( "Could not read config file", e );
		}
		catch ( SAXException e ) {
			throw new ServletException( "Could not parse config file", e );
		}
	}

	private String getContainedText( Node parent, String childTagName ) {
		try {
			Node tag = ( ( Element )parent ).getElementsByTagName( childTagName ).item( 0 );
			String text = ( ( Text )tag.getFirstChild() ).getData();
			return text;
		}
		catch ( Exception e ) {
			return null;
		}
	}

	private void storeDecorator( Decorator d ) {
		decorators.put( d.getName(), d );
	}

	private void storeMapping( String name, String pattern ) {
		mappings.put( pattern, name );
	}

	/** Check if configuration file has been updated, and if so, reload. */
	public synchronized void refresh() throws ServletException {
		if (configLastModified != configFile.lastModified()) loadConfig();
	}

}
